/*
 * An implementation of a monolitic conference protocol.
 *
 * Jan Feenstra, Version 3c (25-02-2002)
 * 
 */

#include "datatype.h"
#include "confprot.h"
#include "processStimuli.h"
#include "processObservation.h"

/*
 * Global Pointer Variable CPD
 *
 */

ConfProtData CPD = &CPV;
static int	stop = 0;

static int	getsocketaddress( char *host_name, int port_number, struct sockaddr_in *addr);
static int	registerUser( char *hostname, int portnumber,  ConfProtData CPD);
static int	init_ConfProt(int argc, char *argv[], ConfProtData CPD);
static void	terminate(int rtnvalue);

int	main(int argc, char *argv[])
{
	int	iohandle = 0;
	fd_set mask;
	MarkCoverage();
	
	/* Set desired termination signals */
	signal(SIGHUP, terminate);	/*  1 */
	signal(SIGINT, terminate);	/*  2 */
	signal(SIGQUIT, terminate);	/*  3 */
	signal(SIGPIPE, terminate);	/* 13 */
	signal(SIGKILL, terminate);	/*  9 */
	signal(SIGTERM, terminate);	/* 15 */
	signal(SIGCHLD, terminate);	/* 18 */

	if ((stop = init_ConfProt(argc, argv, CPD)) != 0) {
		MarkCoverage();
		return(stop);
	}

	MarkCoverage();
#ifdef M57
	(void)raise(SIGQUIT);
#endif

	while (!stop) {
		MarkCoverage();
		FD_ZERO(&mask);
		FD_SET(CPD->sockhandle, &mask);
		FD_SET(CPD->stdinhandle, &mask);
		iohandle = select(32, &mask, (fd_set * ) 0, (fd_set * ) 0, 0);

#ifndef M01
		if (iohandle < 0) {
#endif
		/*
		 * Error in select operation!
		 */
			MarkCoverage();
			return(-1);
#ifndef M01
		}
#endif

		MarkCoverage();
		if (FD_ISSET(CPD->sockhandle, &mask)) {
			MarkCoverage();
			stop = processObservation(CPD);
#ifdef M47
			FlushCache(CPD);
#endif
		} else if (MarkCoverage(),FD_ISSET(CPD->stdinhandle, &mask)) {
			MarkCoverage();
			stop = processStimuli(CPD);
#ifdef M48
			FlushCache(CPD);
#endif
		}

		MarkCoverage();
#ifdef M49
		FlushCache(CPD);
#endif

#ifdef M43
		{
			MarkCoverage();
			int	t = CPD->nickname[MAXNICKNAME-1];
			CPD->nickname[MAXNICKNAME-1] = CPD->nickname[0];
			CPD->nickname[0] = t;
		}
#endif
#ifdef M99
		purecov_save_data();
#endif
	}

	MarkCoverage();
#ifdef M99
	purecov_save_data();
#endif
	return(stop);
}

static int	init_ConfProt(int argc, char *argv[], ConfProtData CPD)
{
	int	i = 0;
	int	mode = 0;
	char hostname[MAXNAME];
	int	portnumber = 0;
	MarkCoverage();

	/*
	 * Setting the default values
	 */
	CPD->gui = 0; /* use USER interface (not the CSAP interface) */
	CPD->debug = 1;
	hostname[0] = '\0';
	CPD->nusers = 1; /* reserve location 0 for the localSocket */
	CPD->availLocalSocket = 0;  /* local socket available? */
	CPD->active = 0;
	CPD->nickname[0] = '\0';
	CPD->conference[0] = '\0';
	CPD->npartners = 0;
	/* CPD->partner[] not initialized! */
	/* CPD->partnernickname[][] not initialized! */

#ifdef MXX
	CPD->nmutants = 0;
	for (i = 0; i < MAXMUTANTS; i++) {
		MarkCoverage();
		CPD->mutant[i].id = 0;
		CPD->mutant[i].Steps = 0;
		CPD->mutant[i].Enable = 0;
		CPD->mutant[i].Disable = -1; /* inactive */
	}
	MarkCoverage();
	i = 0;
#endif

	/* 
	  *Connection to  user.
	  * stdin
	  * stdout/stderr
	  */
	CPD->stdinhandle = 0;
	CPD->stdouthandle = 1;

#ifdef MX1
	/* 
	  * Initialise Cache
	  */
	CPD->nbuckets = 0;
	for (i = 0; i < MAXBUCKETS; i++) {
		MarkCoverage();
		CPD->Cache[i].free = 1;
	}
	MarkCoverage();
	i = 0;
#endif

	for (i = 1; i < argc; i++) {
		MarkCoverage();
		switch (mode) {
		case 0:
			 {
				MarkCoverage();
				if (strcmp(argv[i], "-CONFIG") == 0) {
					MarkCoverage();
					mode = 1;
				} else if (MarkCoverage(), strcmp(argv[i], "-c") == 0) {
					MarkCoverage();
					mode = 1;
				} else if (MarkCoverage(), strcmp(argv[i], "-DEBUG") == 0) {
					MarkCoverage();
					mode = 2;
				} else if (MarkCoverage(), strcmp(argv[i], "-d") == 0) {
					MarkCoverage();
					mode = 2;
				} else if (MarkCoverage(), strcmp(argv[i], "-CSAP") == 0) {
					MarkCoverage();
					CPD->gui = 1;
				} else if (MarkCoverage(), strcmp(argv[i], "-USER") == 0) {
					MarkCoverage();
					CPD->gui = 0;
				}
#ifdef MXX
				else if (MarkCoverage(), strcmp(argv[i], "-MUTANT") == 0) {
					MarkCoverage();
					mode = 3;
				} else if (MarkCoverage(), strcmp(argv[i], "-m") == 0) {
					MarkCoverage();
					mode = 3;
				}
#endif
				else if (MarkCoverage(), strcmp(argv[i], "-a") == 0) {
					MarkCoverage();
					mode = 4;
				} else {
					MarkCoverage();
					mode = 0;
				}
				
				MarkCoverage();
				break;
			}

		case 1:
			 {  /* parse configfile */
				FILE * ifp;
				int	msapid = 0;
				MarkCoverage();

#ifndef M02
				if (((ifp = fopen(argv[i], "r")) == NULL)) {
					MarkCoverage();
#endif
					return(-10);
#ifndef M02
				}
#endif

				while ((CPD->nusers <= MAXUSERS) && (fscanf(ifp, "%d%s%d", &msapid, hostname, &portnumber) == 3)) {
					MarkCoverage();
					if (registerUser(hostname, portnumber, CPD) == 0) {
						MarkCoverage();
						WriteMSG("User %s:%d registered!\n", hostname, portnumber);
					} else
					{
						MarkCoverage();
						WriteMSG("User %s:%d no server!\n", hostname, portnumber);
					}
				}
				MarkCoverage();

				fclose(ifp);

				mode = 0;
				break;
			}

		case 2:
			 { /* assign debug variable */
				MarkCoverage();
				WriteMSG("debug value is %s\n", argv[i]);
				(void)sscanf(argv[i], "%d", &CPD->debug);
				mode = 0;
				break;
			}

#ifdef MXX
		case 3:
			 { /* assign mutant variable */
				int	x; 
				MarkCoverage();

#ifndef M59
				if ((sscanf(argv[i], "%d", &x) == 1)  && (CPD->nmutants < MAXMUTANTS)) {
					MarkCoverage();

					if ((x >= 100) && (x <= 9999)) {
						MarkCoverage();
						CPD->mutant[CPD->nmutants].Enable = x % 100;
						WriteMSG("Register activation delay for mutant %d with value %d\n", x / 100,
						     x % 100);
					} else if (MarkCoverage(), (x >= 10000) && (x <= 19999)) {
						CPD->mutant[CPD->nmutants].Disable = x % 100;
						WriteMSG("Register deactivation delay for mutant %d with value %d\n", (x
						    -10000) / 100, x % 100);
					} else {
						MarkCoverage();
						CPD->mutant[CPD->nmutants].id = x;
						WriteMSG("Mutant %d registered!\n", CPD->mutant[CPD->nmutants].id);
						(CPD->nmutants)++;
					}
					MarkCoverage();

				} else
#endif
				{
					MarkCoverage();
					WriteMSG("Mutant with value %s skipped!\n", argv[i]);
				}
				mode = 0;
				break;
			}
#endif

		case 4:
			 {
				int	n = 0;
				MarkCoverage();

				/*
				 * Determine specified network adress!
				 * "pythagoras:1025" where pythagoras is the hostname 
				 * and the positive number 1025 is the port address.
				 */
				n = (int)strlen(argv[i]);
				while ((n > 1) && isdigit((int)argv[i][n-1])) {
					MarkCoverage();
					n--;
				}
				MarkCoverage();
				strncpy(hostname, argv[i], n - 1);
				hostname[n-1] = '\0';
				(void)sscanf(&argv[i][n], "%d", &portnumber);

				/*
				 * Add potential user, if possible
				 */

				MarkCoverage();
				if (registerUser(hostname, portnumber, CPD) == 0) {
					MarkCoverage();
					WriteMSG("User %s:%d registered!\n", hostname, portnumber);
				} else
				 {
					MarkCoverage();
					WriteMSG("User %s:%d no server!\n", hostname, portnumber);
				}

				MarkCoverage();
				mode = 0;
				break;
			}
		}
		MarkCoverage();
	}
	MarkCoverage();
	
	/* 
	  * Do we have a local socket?
	  */
	if (CPD->nusers == 0) {
		MarkCoverage();
		return(-11);
	}

	MarkCoverage();

#ifdef MXX
	WriteMSG("%d mutants are registered!\n", CPD->nmutants);
#endif

	return(0);
}


static int	getsocketaddress( char *host_name, int port_number, struct sockaddr_in *addr)
{
	struct hostent *host;
	MarkCoverage();

#ifndef M04
	if (((host = gethostbyname(host_name)) == 0)) {
		MarkCoverage();
#endif
		return(-1);
#ifndef M04
	} else {
		MarkCoverage();
		memcpy((char *) & addr->sin_addr, (char *) host->h_addr, host->h_length);
		addr->sin_family = AF_INET;
		addr->sin_port = htons(port_number);
		return(0);
	}
#endif
}


static int	registerUser( char *hostname, int portnumber,  ConfProtData CPD)
{
	/*
	 * Leave if maximal number of user reached!
	 */
	MarkCoverage();
#ifndef M03
	if ((CPD->nusers >= MAXUSERS)) {
		MarkCoverage();
#endif
		return(-2);
#ifndef M03
	}
		MarkCoverage();
#endif

	/*
	 * Convert hostname and portaddress to machine socketadress
	 */
	if (getsocketaddress(hostname, portnumber, &CPD->socketaddress[CPD->nusers])) {
		MarkCoverage();
		return(-1); /* Error in networkaddress */
	}
	MarkCoverage();

	if (CPD->availLocalSocket == 1) {
		/*
		 * Add user to list. 
		 */
		MarkCoverage();
		CPD->nusers++;
		return(0);
	}
	MarkCoverage();

	/*
	  * Create sockethandle 
	  */
#ifndef M05
	if (((CPD->sockhandle = socket(AF_INET, SOCK_DGRAM, 0)) < 0 )) {
		MarkCoverage();
#endif
		CPD->nusers++;
		return(-2); /* Error during sockethandle creation */
#ifndef M05
	}
	MarkCoverage();
#endif

	/*
	  * Bind sockethandle to networkaddress
	  */
#ifndef M56
	if ((bind(CPD->sockhandle,  (struct sockaddr *) & CPD->socketaddress[CPD->nusers],  sizeof(struct sockaddr_in )) < 0)){
	MarkCoverage();
#endif
		CPD->nusers++;
		return(-3); /* socket not bindable to network address! */
#ifndef M56
	}
	MarkCoverage();
#endif

	/*
	 * Put server at 0th entry
	 */
	memcpy(&CPD->socketaddress[0], &CPD->socketaddress[CPD->nusers], sizeof(struct sockaddr_in ));
	CPD->availLocalSocket = 1;
	
	return(0);
}


void	transmit(ConfProtData CPD,  char *pdu, int pdusize, int typ, struct sockaddr_in *addr)
{
	MarkCoverage();
#ifndef M46
		sendto(CPD->sockhandle, pdu, pdusize, 0, \
    	(struct sockaddr *)addr, sizeof(struct sockaddr_in ));
#endif
#ifdef M46
#ifndef M58
	if ((CPD->nbuckets == MAXBUCKETS)) {
		MarkCoverage();
#endif
		FlushCache(CPD);
#ifndef M58
	}
	MarkCoverage();
#endif

	WriteMSG("Caching outgoing pdu!\n");
	CPD->Cache[CPD->nbuckets].sort = typ;
	memcpy(CPD->Cache[CPD->nbuckets].buf, pdu, pdusize);
	memcpy(&CPD->Cache[CPD->nbuckets].address, addr, sizeof(struct sockaddr_in ));
	CPD->Cache[CPD->nbuckets].sizeofbuf = pdusize;
	CPD->Cache[CPD->nbuckets].free = 0;
	CPD->nbuckets++;
#endif
}

#ifdef MX1
void	FlushCache(ConfProtData CPD)
{
	int	i = 0;

	MarkCoverage();
	WriteMSG("Flushing cache!\n");
	for (i = 0; i < CPD->nbuckets; i++) {
		MarkCoverage();
		sendto(CPD->sockhandle, CPD->Cache[i].buf, CPD->Cache[i].sizeofbuf, 0,  (struct sockaddr *) & CPD->Cache[i].address,
		     sizeof(struct sockaddr_in ));
		CPD->Cache[i].free = 1;
	}
	MarkCoverage();
	CPD->nbuckets = 0;
}
#endif MX1

static void	terminate(int rtnvalue)
{
	MarkCoverage();
	WriteMSG("A Signal(%d) detected!\n", rtnvalue);
	stop = -2;
}

